import datetime as dt
import pathlib
import unittest
from unittest.mock import DEFAULT, patch

from click.testing import CliRunner

import clishelf.emoji as emoji


def side_effect_func(*args, **kwargs):
    print("----------", args, kwargs)
    if "emoji.py" in args[0]:
        _ = kwargs
        return pathlib.Path(__file__)
    return DEFAULT


class CLIEmojiTestCase(unittest.TestCase):
    def setUp(self) -> None:
        self.maxDiff = None
        self.runner = CliRunner()

    @patch("clishelf.emoji.Path", side_effect=side_effect_func)
    @patch("clishelf.emoji.datetime")
    @patch("clishelf.emoji.requests.get")
    def test_fetch_emoji(self, mock_request, mock_now, mock_path):
        mock_request.get.return_value.json.return_value = []
        mock_now.now.return_value = dt.datetime(2024, 1, 1, 0, 0, 0)
        result = self.runner.invoke(emoji.fetch)
        self.assertTrue(mock_path.called)
        self.assertEqual(0, result.exit_code)

        test_file = pathlib.Path(__file__).parent / "assets/emoji.json"
        self.assertTrue(test_file.exists())

        result = self.runner.invoke(emoji.fetch, args="-b")
        self.assertEqual(0, result.exit_code)

        test_file_bk = (
            pathlib.Path(__file__).parent / "assets/emoji.bk20240101000000.json"
        )
        self.assertTrue(test_file_bk.exists())
        test_file.unlink()
        test_file_bk.unlink()
        test_file_bk.parent.rmdir()



import unittest
from datetime import datetime
from unittest.mock import patch

import clishelf.settings as settings


class SettingsTestCase(unittest.TestCase):
    def test_bump_version_setting(self):
        bump_setting = settings.BumpVerConf

        self.assertEqual(bump_setting.regex, bump_setting.get_regex())
        self.assertEqual(
            bump_setting.regex_dt,
            bump_setting.get_regex(is_dt=True),
        )

        update_dt_pre = bump_setting.update_dt_pre
        self.assertEqual("20240102.1", update_dt_pre("20240102"))
        self.assertEqual("20240102.6", update_dt_pre("20240102.5"))
        with self.assertRaises(ValueError):
            bump_setting.update_dt_pre("202401.post")

    def test_bump_version_get_version(self):
        bump_setting = settings.BumpVerConf
        rs = bump_setting.get_version(
            version=3,
            params={
                "version": "",
                "changelog": "",
                "file": "",
            },
        )
        self.assertEqual(
            rs,
            bump_setting.v1.format(
                changelog="",
                main=bump_setting.main.format(
                    version="",
                    msg=bump_setting.msg,
                    regex=bump_setting.regex,
                    file="",
                ),
            ),
        )

    @patch("clishelf.settings.datetime.datetime")
    def test_bump_version_get_version_dt(self, mock_datetime):
        mock_datetime.now.return_value = datetime(2024, 1, 1, 0, 0)
        bump_setting = settings.BumpVerConf
        rs = bump_setting.get_version(
            version=2,
            params={
                "version": "",
                "changelog": "",
                "file": "",
                "action": "date",
            },
            is_dt=True,
        )
        self.assertEqual(
            rs,
            bump_setting.v2.format(
                changelog="",
                main=bump_setting.main_dt.format(
                    version="",
                    new_version="20240101",
                    msg=bump_setting.msg,
                    regex=bump_setting.regex_dt,
                    file="",
                ),
            ),
        )

        rs = bump_setting.get_version(
            version=2,
            params={
                "version": "20240101",
                "changelog": "",
                "file": "",
                "action": "pre",
            },
            is_dt=True,
        )
        self.assertEqual(
            rs,
            bump_setting.v2.format(
                changelog="",
                main=bump_setting.main_dt.format(
                    version="20240101",
                    new_version="20240101.1",
                    msg=bump_setting.msg,
                    regex=bump_setting.regex_dt,
                    file="",
                ),
            ),
        )

        with self.assertRaises(ValueError):
            bump_setting.get_version(
                version=2,
                params={
                    "version": "20240101",
                    "changelog": "",
                    "file": "",
                    "action": "post",
                },
                is_dt=True,
            )



# ------------------------------------------------------------------------------
# Copyright (c) 2022 Korawich Anuttra. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root for
# license information.
# ------------------------------------------------------------------------------
from __future__ import annotations

import datetime
import re
from textwrap import dedent


class GitConf:
    """Git Config."""

    branch_types: list[str] = ["feature", "bug", "hot"]

    # These branch names are not validated with this same rules
    # (permissions should be configured on the server if you want to prevent
    # pushing to any of these):
    branch_excepts: list[str] = [
        "feature",
        "dev",
        "main",
        "stable",
        # for quickly fixing critical issues, usually with a temporary solution.
        "hotfix",
        "bugfix",  # for fixing a bug
        "feature",  # for adding, removing or modifying a feature
        "test",  # for experimenting something which is not an issue
        "wip",  # for a work in progress
    ]

    regex_branch_types: str = "|".join(branch_types)

    regex_commit_msg: str = (
        r"(?P<prefix>\w+)(?:\((?P<topic>\w+)\))?: (?P<header>.+)"
    )

    # TODO: reference emoji from https://gitmoji.dev/
    #   All emojis, https://github.com/ikatyang/emoji-cheat-sheet
    #   GitHub API: https://api.github.com/emojis
    commit_prefix: tuple[tuple[str, str, str]] = (
        ("feature", "Features", ":dart:"),  # 🎯, 📋 :clipboard:, ✨ :sparkles:
        ("feat", "Features", ":dart:"),  # 🎯, 📋 :clipboard:, ✨ :sparkles:
        ("hotfix", "Fix Bugs", ":fire:"),  # 🔥, 🚑 :ambulance:
        ("fixed", "Fix Bugs", ":gear:"),  # ⚙️, 🛠️ :hammer_and_wrench:
        ("fix", "Fix Bugs", ":gear:"),  # ⚙️, 🛠️ :hammer_and_wrench:
        ("bug", "Fix Bugs", ":bug:"),  # 🐛
        ("docs", "Documents", ":page_facing_up:"),  # 📄, 📑 :bookmark_tabs:
        ("styled", "Code Changes", ":art:"),  # 🎨, 📝 :memo:, ✒️ :black_nib:
        ("style", "Code Changes", ":art:"),  # 🎨, 📝 :memo:, ✒️ :black_nib:
        ("refactored", "Code Changes", ":construction:"),
        # 🚧, 💬 :speech_balloon:
        ("refactor", "Code Changes", ":construction:"),
        # 🚧, 💬 :speech_balloon:
        ("perf", "Code Changes", ":zap:"),
        # ⚡, 📈 :chart_with_upwards_trend:, ⌛ :hourglass:
        ("tests", "Code Changes", ":test_tube:"),  # 🧪, ⚗️ :alembic:
        ("test", "Code Changes", ":test_tube:"),  # 🧪, ⚗️ :alembic:
        ("build", "Build & Workflow", ":toolbox:"),  # 🧰, 📦 :package:
        ("workflow", "Build & Workflow", ":rocket:"),  # 🚀, 🕹️ :joystick:
        ("deps", "Dependencies", ":pushpin:"),  # 📌, 🔍 :mag:
        ("dependency", "Dependencies", ":pushpin:"),  # 📌, 🔍 :mag:
        ("secure", "Security", ":lock:"),  # 🔒
        ("init", "Features", ":tada:"),  # 🎉
        ("deprecate", "Code Changes", ":wastebasket:"),  # 🗑️
        ("revert", "Code Changes", ":rewind:"),  # ⏪
    )

    commit_prefix_group: tuple[tuple[str, str]] = (
        ("Features", ":sparkles:"),  # ✨
        ("Code Changes", ":black_nib:"),  # ✒️
        ("Documents", ":card_file_box:"),  # 🗃️, 📑 :bookmark_tabs:
        ("Fix Bugs", ":bug:"),  # 🐛
        ("Build & Workflow", ":package:"),  # 📦
        ("Dependencies", ":postbox:"),  # 📮
        ("Security", ":closed_lock_with_key:"),  # 🔐
    )


class BumpVerConf:
    """Bump Version Config."""

    main: str = dedent(
        r"""
    [bumpversion]
    current_version = {version}
    commit = True
    tag = False
    parse = ^
        {regex}
    serialize =
        {{major}}.{{minor}}.{{patch}}.{{prekind}}{{pre}}.{{postkind}}{{post}}
        {{major}}.{{minor}}.{{patch}}.{{prekind}}{{pre}}
        {{major}}.{{minor}}.{{patch}}.{{postkind}}{{post}}
        {{major}}.{{minor}}.{{patch}}
    message = {msg}

    [bumpversion:part:prekind]
    optional_value = _
    values =
        _
        a
        b
        rc

    [bumpversion:part:postkind]
    optional_value = _
    values =
        _
        post

    [bumpversion:file:{file}]
    """
    ).strip()

    main_dt: str = dedent(
        r"""
    [bumpversion]
    current_version = {version}
    new_version = {new_version}
    commit = True
    tag = False
    parse = ^
        {regex}
    serialize =
        {{date}}.{{pre}}
        {{date}}
    message = {msg}

    [bumpversion:file:{file}]
    """
    )
    # main_dt: str = dedent(
    #     r"""
    # [bumpversion]
    # current_version = {version}
    # new_version = {new_version}
    # commit = True
    # tag = False
    # parse = ^
    #     {regex}
    # serialize =
    #     {{date}}.{{pre}}
    #     {{date}}
    # message = {msg}
    #
    # [bumpversion:file:{file}]
    # """
    # )

    msg: str = (
        # 🔖 :bookmark:
        ":bookmark: Bump up to version {current_version} -> {new_version}."
    )

    regex: str = (
        r"(?P<major>\d+)\.(?P<minor>\d+)\.(?P<patch>\d+)"
        r"(\.(?P<prekind>a|alpha|b|beta|d|dev|rc)(?P<pre>\d+))?"
        r"(\.(?P<postkind>post)(?P<post>\d+))?"
    )

    regex_dt: str = r"(?P<date>\d{4}\d{2}\d{2})(\.(?P<pre>\d+))?"

    v1: str = dedent(
        r"""
    {main}

    [bumpversion:file:{changelog}]
    search = {{#}}{{#}} Latest Changes
    replace = {{#}}{{#}} Latest Changes

        {{#}}{{#}} {{new_version}}
    """
    ).strip()

    v2: str = dedent(
        r"""
    {main}

    [bumpversion:file:{changelog}]
    search = {{#}}{{#}} Latest Changes
    replace = {{#}}{{#}} Latest Changes

        {{#}}{{#}} {{new_version}}

        Released: {{utcnow:%Y-%m-%d}}
    """
    ).strip()

    @classmethod
    def get_version(
        cls,
        version: int,
        params: dict[str, str],
        is_dt: bool = False,
    ):
        """Generate the `bump2version` config from specific version"""
        if not hasattr(cls, f"v{version}"):
            version = 1
        template: str = getattr(cls, f"v{version}")
        if is_dt:
            if (action := params.get("action", "date")) == "date":
                new_version: str = datetime.datetime.now().strftime("%Y%m%d")
            elif action == "pre":
                new_version = cls.update_dt_pre(params.get("version"))
            else:
                raise ValueError(
                    f"the action does not support for {action} with use "
                    f"datetime mode."
                )
            return template.format(
                changelog=params.get("changelog"),
                main=cls.main_dt.format(
                    version=params.get("version"),
                    new_version=new_version,
                    msg=cls.msg,
                    regex=cls.regex_dt,
                    file=params.get("file"),
                ),
            )
        return template.format(
            changelog=params.get("changelog"),
            main=cls.main.format(
                version=params.get("version"),
                msg=cls.msg,
                regex=cls.regex,
                file=params.get("file"),
            ),
        )

    @classmethod
    def update_dt_pre(cls, version: str):
        """Return new pre version of datetime mode.
        Examples:
            20240101        ->  20240101.1
            20240101.2      ->  20240101.3
            20240101.post   ->  20240101.1
        """
        if search := re.search(BumpVerConf.regex_dt, version):
            search_dict: dict[str, str] = search.groupdict()
            if pre := search_dict.get("pre"):
                pre = str(int(pre) + 1)
            else:
                pre = "1"
            return f"{search_dict['date']}.{pre}"
        raise ValueError(
            "version value does not match with datetime regex string."
        )

    @classmethod
    def get_regex(cls, is_dt: bool = False) -> str:
        return cls.regex_dt if is_dt else cls.regex



import datetime
import sys
import unittest
from unittest.mock import DEFAULT, patch

import clishelf.git as git


def side_effect_func(*args, **kwargs):
    if any(["git", "config", "--local", "user.name"] == a for a in args):
        _ = kwargs
        return "Test User".encode(encoding=sys.stdout.encoding)
    elif any(["git", "config", "--local", "user.email"] == a for a in args):
        _ = kwargs
        return "test@mail.com".encode(encoding=sys.stdout.encoding)
    else:
        return DEFAULT


class GitModelTestCase(unittest.TestCase):
    def test_commit_prefix_model(self):
        rs = git.CommitPrefix(
            name="test",
            group="A",
            emoji=":dart:",
        )
        self.assertEqual(hash(rs), hash(rs.name))
        self.assertEqual("test", str(rs))

    def test_commit_prefix_group_model(self):
        rs = git.CommitPrefixGroup(
            name="test",
            emoji=":dart:",
        )
        self.assertEqual(hash(rs), hash(rs.name))
        self.assertEqual("test", str(rs))

    def test_commit_message_model(self):
        msg = git.CommitMsg(
            content=":dart: feat: start initial testing",
            mtype=None,
        )
        self.assertEqual(
            "Features: :dart: feat: start initial testing", str(msg)
        )

        msg = git.CommitMsg(
            content=":dart: demo: start initial testing",
            mtype=None,
        )
        self.assertEqual(
            "Code Changes: :dart: demo: start initial testing", str(msg)
        )

        msg = git.CommitMsg(
            content=":dart: start initial testing",
            mtype=None,
        )
        self.assertEqual("Code Changes: :dart: start initial testing", str(msg))

        msg = git.CommitMsg(
            content="⬆️ deps: upgrade dependencies from main branch (#63)",
            mtype=None,
        )
        self.assertEqual(
            (
                "Dependencies: :arrow_up: deps:  upgrade dependencies from "
                "main branch (#63)"
            ),
            str(msg),
        )

    def test_commit_message_model_raise(self):
        with self.assertRaises(ValueError):
            git.CommitMsg(
                content="demo: start initial testing",
                mtype=None,
            )

    @patch("clishelf.utils.load_pyproject")
    def test_commit_message_model_raise_without_conf(self, mock_load_pyproject):
        with self.assertRaises(ValueError):
            mock_load_pyproject.return_value = {
                "tool": {"shelf": {"git": {"commit_prefix_force_fix": False}}},
            }
            git.CommitMsg(
                content="⬆️ demo: start initial testing",
                mtype=None,
            )

    def test_commit_log_model(self):
        log = git.CommitLog(
            hash="test",
            refs="HEAD",
            date=datetime.date(2023, 1, 1),
            msg=git.CommitMsg(":dart: feat: start initial testing"),
            author=git.Profile(
                name="test",
                email="test@mail.com",
            ),
        )
        self.assertEqual(
            (
                "test|2023-01-01|:dart: feat: start initial testing|"
                "test|test@mail.com|HEAD"
            ),
            str(log),
        )


class GitTestCase(unittest.TestCase):

    @patch("clishelf.git.subprocess.check_output", side_effect=side_effect_func)
    @patch("clishelf.utils.load_pyproject")
    def test_load_profile(self, mock_load_pyproject, mock):
        mock_load_pyproject.return_value = {}
        rs = git.load_profile()
        self.assertIsInstance(rs, git.Profile)
        self.assertEqual("Test User", rs.name)
        self.assertEqual("test@mail.com", rs.email)
        self.assertTrue(mock.called)

    def test_get_commit_prefix(self):
        data = git.get_commit_prefix()

        # This assert will true if run on `pytest -v`
        self.assertEqual(24, len(data))

    def test_get_commit_prefix_group(self):
        data: tuple[git.CommitPrefixGroup] = git.get_commit_prefix_group()
        feat: git.CommitPrefixGroup = [
            cm for cm in data if cm.name == "Features"
        ][0]
        self.assertEqual(":tada:", feat.emoji)

    def test_get_branch_name(self): ...

    def test_get_latest_tag(self): ...

    def test_gen_commit_log(self): ...

    def test_git_demojize(self):
        self.assertEqual(
            "test :fire: :fire:",
            git.git_demojize("test 🔥 :fire:"),
        )

    def test_validate_commit_msg_warning(self):
        rs = git._validate_commit_msg_warning([":dart: feat: demo", ""])
        self.assertListEqual(
            rs,
            [
                (
                    "There should be between 21 and 50 characters in the "
                    "commit title."
                ),
                "There should at least 3 lines in your commit message.",
                "There should not has dot in the end of commit message.",
            ],
        )

        rs = git._validate_commit_msg_warning(
            [":dart: feat: demo test validate for warning.", "empty"]
        )
        self.assertListEqual(
            rs,
            [
                "There should at least 3 lines in your commit message.",
                (
                    "There should be an empty line between the commit title "
                    "and body."
                ),
            ],
        )

        rs = git._validate_commit_msg_warning(
            [
                ":dart: feat: demo test validate for warning.",
                "",
                "body of commit log",
                "",
            ]
        )
        self.assertListEqual(
            rs,
            [],
        )

    def test_validate_commit_msg(self):
        rs = git.validate_commit_msg([])
        self.assertTupleEqual(
            rs,
            (
                ["Please supply commit message without start with ``#``."],
                git.Level.ERROR,
            ),
        )

        rs = git.validate_commit_msg(
            [
                ":dart: feat: demo test validate for warning.",
                "",
                "body of commit log",
                "",
            ]
        )
        self.assertTupleEqual(
            rs,
            (
                ["The commit message has the required pattern."],
                git.Level.OK,
            ),
        )

        rs = git.validate_commit_msg(
            [
                ":dart: feat: demo test validate for warning.",
                "",
                (
                    "body of commit log that has character more that 72 and "
                    "it will return some warning message from function"
                ),
                "",
            ]
        )
        self.assertTupleEqual(
            rs,
            (
                ["The commit body should wrap at 72 characters at line: 3."],
                git.Level.WARNING,
            ),
        )
